function [xopt,fopt,out]=glis(f0,lb,ub,opts)
% GLIS Solve (GL)obal optimization problems using (I)nverse distance
% weighting and radial basis function (S)urrogates.
%
% (C) 2019 A. Bemporad, June 14, 2019
%
% [xopt,fopt,out] = glis(f,lb,ub,opts) solves the global optimization problem 
%
% min  f(x)
% s.t. lb <= x <=ub, A*x <=b, g(x)<=0
%
% using the global optimization algorithm described in [1]. The approach is
% particularly useful when f(x) is time-consuming to evaluate, as it
% attempts at minimizing the number of function evaluations.
%
% The input argument opts specifies various parameters of the algorithm:
%
% opts.maxevals:     maximum number of function evaluations
% opts.alpha:        weight on function uncertainty variance measured by IDW
% opts.delta:        weight on distance from previous samples
% opts.nsamp:        number of initial samples
% opts.useRBF:       true = use RBFs, false = use IDW interpolation
% opts.rbf:          function handle to RBF function (only used if opts.useRBF=true)
% opts.scalevars:    scale problem variables (default: true)
% opts.svdtol:       tolerance used to discard small singular values
% opts.Aineq:        matrix A defining linear inequality constraints 
% opts.bineq:        right hand side of constraints A*x <= b
% opts.g:            constraint function handle
% opts.shrink_range  if 0, disable shrinking lb and ub to bounding box of feasible set
% opts.constraint_penalty: penalty term on violation of linear inequality
%                          and nonlinear constraints
% opts.feasible_sampling: if true, initial samples are forced to be feasible 
% opts.epsDeltaF:    minimum value used to scale the IDW distance function
% opts.globoptsol:   nonlinear solver used during acquisition.
%                    interfaced solvers are:
%                      'direct' DIRECT from NLopt tool (nlopt.readthedocs.io)
%                      'pswarm' PSWarm solver v2.1 (www.norg.uminho.pt/aivaz/pswarm/)
%                      'tmw-pso' Particle Swarm Optimizer from MATLAB Global Optimization Toolbox
%                      'tmw-ga' Genetic Algorithm from MATLAB Global Optimization Toolbox
% opts.PSOiters:     number of iterations in PSO solver
% opts.PSOwarmsize:  swarm size in PSO solver
% opts.display:      verbosity level (0=minimum)
%    
% The output argument 'out' is a structure reporting the following information:
%
% out.X:  trace of all samples x at which f(x) has been evaluated
% out.F:  trace of all function evaluations f(x)
% out.W:  final set of weights (only meaningful for RBFs)
% out.M   RBF matrix (only meaningful for RBFs)
% out.xopt: best sample found during search
% out.fopt: best value found during search, fopt=f(xopt)
%
% [1] A. Bemporad, "Global optimization via inverse weighting and radial 
% basis functions", arXiv:1906.06498v1, June 15, 2019.
% https://arxiv.org/pdf/1906.06498.pdf

[nvar,Aineq,bineq,g,isLinConstrained,isNLConstrained,...
    X,F,z,nsamp,maxevals,epsDeltaF,alpha,delta,rhoC,display,svdtol,...
    dd,d0,useRBF,rbf,M,scalevars,globoptsol,pswarm_vars,direct_vars]=...
    glis_init(lb,ub,opts);

if scalevars
    f=@(X) f0(X.*(ones(size(X,1),1)*dd')+ones(size(X,1),1)*d0');
else
    f=f0;
end

if ~display
    F(1:nsamp)=f(X(1:nsamp,:));
else
    fprintf('Generating initial samples:\n');
    for i=1:nsamp
        F(i)=f(X(i,:));
        fprintf('N = %4d, cost = %8g. ',i,F(i));
        for j=1:nvar
            aux=X(i,j);
            if scalevars
                aux=aux*dd(j)+d0(j);
            end
            fprintf('x%d = %5.4f ',j,aux);
        end
        fprintf('\n');
    end
    fprintf('\nDone. Now starting optimizing:\n\n');
end

if useRBF
    W=get_rbf_weights(M,F,nsamp,svdtol);
else
    W=[];
end

fbest=Inf;
zbest=zeros(nsamp,1);
for i=1:nsamp
    isfeas=true;
    if isLinConstrained
        isfeas=isfeas && all(Aineq*X(i,:)'<=bineq);
    end
    if isNLConstrained
        isfeas=isfeas && all(g(X(i,:)')<=0);
    end
    if isfeas && fbest>F(i)
        fbest=F(i);
        zbest=X(i,:)';
    end
end

Fmax=max(F(1:nsamp));
Fmin=min(F(1:nsamp));

N=nsamp;
while N<maxevals

    dF=max(Fmax-Fmin,epsDeltaF);
    
    if isLinConstrained || isNLConstrained
        penalty=rhoC*dF;
    end
    if isLinConstrained && isNLConstrained
        constrpenalty=@(x) penalty*(sum(max(Aineq*x(:)-bineq,0).^2) +...
            sum(max(g(x(:)),0).^2));
    elseif isLinConstrained && ~isNLConstrained
        constrpenalty=@(x) penalty*(sum(max(Aineq*x(:)-bineq,0).^2));
    elseif ~isLinConstrained && isNLConstrained
        constrpenalty=@(x) penalty*sum(max(g(x(:)),0).^2);
    else
        constrpenalty=@(x) 0;
    end

    acquisition=@(x,p) facquisition(x(:)',X,F,N,alpha,delta,dF,W,rbf) + ...
        constrpenalty(x);
    
    switch globoptsol
        case 'pswarm'
            pswarm_vars.Problem.ObjFunction= @(x) facquisition(x(:)',X,F,N,alpha,delta,dF,W,rbf)+...
                constrpenalty(x(:));
            evalc('z=PSwarm(pswarm_vars.Problem,pswarm_vars.InitialPopulation,pswarm_vars.Options);');
            
        case 'direct'
            direct_vars.opt.min_objective = acquisition;
            zold=z;
            z=nlopt_optimize(direct_vars.opt,zold);
            z=z(:);
        
		case {'tmw-pso','tmw-ga'}
            lb2=lb;
            ub2=ub;
            if scalevars
                lb2=-ones(nvar,1);
                ub2=ones(nvar,1);
            end
            if strcmp(globoptsol,'tmw-pso')
                z=particleswarm(acquisition,nvar,lb2,ub2,pswarm_vars.Options);
            else
                z=ga(acquisition,nvar,[],[],[],[],lb2,ub2,[],pswarm_vars.Options);
            end
            z=z(:);
    end
    
    fz = f(z'); % function evaluation
    
    N=N+1;

    X(N,:)=z';
    F(N)=fz;
    
    Fmax=max(Fmax,fz);
    Fmin=min(Fmin,fz);
    
    if useRBF
        % Just update last row and column of M
        for h=1:N
            mij=rbf(X(h,:),X(N,:));
            M(h,N)=mij;
            M(N,h)=mij;
        end
        W=get_rbf_weights(M,F,N,svdtol);
    end
    
    if fbest>fz
        fbest=fz;
        zbest=z;
    end
    
    if display
        fprintf('N = %4d, cost = %8g, best = %8g. ',N,fz,fbest);
        for j=1:nvar
            aux=zbest(j);
            if scalevars
                aux=aux*dd(j)+d0(j);
            end
            fprintf('x%d = %5.4f ',j,aux);
        end
        fprintf('\n');
    end
end

xopt=zbest;
if scalevars
    % Scale variables back
    xopt=xopt.*dd+d0;
    X=X.*(ones(N,1)*dd')+ones(N,1)*d0';
end

fopt=fbest;

if ~useRBF
    W=[];
end

out=struct('X',X,'F',F,'W',W,'M',M,'fopt',fopt,'xopt',xopt);

%%%%%%%%%%%%%%%%%%%%%%
function [f,fhat,dhat]=facquisition(x,X,F,N,alpha,delta,dF,W,rbf)
% Acquisition function to minimize to get next sample

m=size(x,1); % number of points x to evaluate the acquisition function

f=zeros(m,1);

for i=1:m
    xx=x(i,:)';
    
    d=sum((X(1:N,:)-ones(N,1)*xx').^2,2);
    ii=find(d<1e-12);
    if ~isempty(ii)
        fhat=F(ii(1));
        dhat=0;
    else
        w=exp(-d)./d;
        sw=sum(w);
        
        if ~isempty(rbf)
            v=zeros(N,1);
            for j=1:N
                v(j)=rbf(X(j,:),xx');
            end
            fhat=v'*W;
        else
            fhat=sum(F(1:N).*w)/sw;
        end
        
        dhat=delta*atan(1/sum(1./d))*2/pi*dF+...
            alpha*sqrt(sum(w.*(F(1:N)-fhat).^2)/sw);
    end
    f(i)=fhat-dhat;
end


%%%%%%%%%%%%%%%%%%%%%%
function W=get_rbf_weights(M,F,NX,svdtol)
% Solve M*W = F using SVD
        
[U,S,V]=svd(M(1:NX,1:NX));
dS=diag(S);
ns=find(dS>=svdtol,1,'last');
W=V(:,1:ns)*diag(1./dS(1:ns))*U(:,1:ns)'*F(1:NX);
